# AUTOGENERATED! DO NOT EDIT! File to edit: index.ipynb (unless otherwise specified).

__all__ = []

# Cell
from .data import *
from .measures import *
from .statistics import *
from .labelling import *
from .machine_learning import *


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"read_csv": "20_data.ipynb",
         "concat": "20_data.ipynb",
         "add_tables": "22_statistics.ipynb",
         "prepare_labrie_data": "20_data.ipynb",
         "prepare_braverman_data": "20_data.ipynb",
         "prepare_philander_data": "20_data.ipynb",
         "generate_bets": "20_data.ipynb",
         "check_data": "20_data.ipynb",
         "summarise_app": "20_data.ipynb",
         "summarise_providers": "20_data.ipynb",
         "plot_player_career": "20_data.ipynb",
         "plot_player_career_split": "20_data.ipynb",
         "plot_provider_dates": "20_data.ipynb",
         "plot_provider_bets": "20_data.ipynb",
         "create_measures_table": "21_measures.ipynb",
         "duration": "21_measures.ipynb",
         "frequency": "21_measures.ipynb",
         "frequency_aggregated": "21_measures.ipynb",
         "frequency_percent": "21_measures.ipynb",
         "number_of_bets": "21_measures.ipynb",
         "mean_bets_per_day_aggregated": "21_measures.ipynb",
         "mean_bets_per_day": "21_measures.ipynb",
         "median_bets_per_day": "21_measures.ipynb",
         "num_sessions": "21_measures.ipynb",
         "total_play_time": "21_measures.ipynb",
         "mean_sessions_per_day": "21_measures.ipynb",
         "median_sessions_per_day": "21_measures.ipynb",
         "mean_session_duration": "21_measures.ipynb",
         "median_session_duration": "21_measures.ipynb",
         "mean_play_time": "21_measures.ipynb",
         "median_play_time": "21_measures.ipynb",
         "num_bets_deviation_per_day": "21_measures.ipynb",
         "inactive_day_streak_variance": "21_measures.ipynb",
         "persistence": "21_measures.ipynb",
         "bet_count_trajectory": "21_measures.ipynb",
         "mean_bets_per_session": "21_measures.ipynb",
         "median_bets_per_session": "21_measures.ipynb",
         "bets_per_session_variance": "21_measures.ipynb",
         "session_duration_variance": "21_measures.ipynb",
         "session_duration_trajectory": "21_measures.ipynb",
         "sawtooth_occurances": "21_measures.ipynb",
         "active_day_trajectory": "21_measures.ipynb",
         "total_wagered": "21_measures.ipynb",
         "mean_bet_size": "21_measures.ipynb",
         "median_bet_size": "21_measures.ipynb",
         "max_bet": "21_measures.ipynb",
         "min_bet": "21_measures.ipynb",
         "bet_size_range": "21_measures.ipynb",
         "bet_size_deviation": "21_measures.ipynb",
         "trajectory": "21_measures.ipynb",
         "mean_amount_wagered_per_day": "21_measures.ipynb",
         "median_amount_wagered_per_day": "21_measures.ipynb",
         "amount_wagered_per_day_variance": "21_measures.ipynb",
         "mean_amount_wagered_per_session": "21_measures.ipynb",
         "median_amount_wagered_per_session": "21_measures.ipynb",
         "total_amount_wagered_across_duration": "21_measures.ipynb",
         "amount_wagered_per_session_variance": "21_measures.ipynb",
         "bet_size_trajectory": "21_measures.ipynb",
         "net_loss": "21_measures.ipynb",
         "percent_loss": "21_measures.ipynb",
         "payout_bet_count_ratio": "21_measures.ipynb",
         "payout_bet_size_ratio": "21_measures.ipynb",
         "mean_net_loss_per_session": "21_measures.ipynb",
         "median_net_loss_per_session": "21_measures.ipynb",
         "sum_of_payouts": "21_measures.ipynb",
         "relative_big_win": "21_measures.ipynb",
         "big_win": "21_measures.ipynb",
         "mean_payout_size": "21_measures.ipynb",
         "median_payout_size": "21_measures.ipynb",
         "overall_binary_loser": "21_measures.ipynb",
         "clamped_net_win": "21_measures.ipynb",
         "net_loss_on_last_day": "21_measures.ipynb",
         "mean_loss_per_bet": "21_measures.ipynb",
         "median_loss_per_bet": "21_measures.ipynb",
         "theoretical_loss": "21_measures.ipynb",
         "mean_odds_per_bet": "21_measures.ipynb",
         "median_odds_per_bet": "21_measures.ipynb",
         "num_providers_used": "21_measures.ipynb",
         "mean_game_types_per_day": "21_measures.ipynb",
         "calculate_labrie_measures": "21_measures.ipynb",
         "calculate_braverman_measures": "21_measures.ipynb",
         "available_measures": "21_measures.ipynb",
         "check_measure_data": "21_measures.ipynb",
         "get_sessions": "21_measures.ipynb",
         "get_daily_bets": "21_measures.ipynb",
         "standardise_measures_table": "21_measures.ipynb",
         "split_measures_table": "21_measures.ipynb",
         "generic_trajectory": "21_measures.ipynb",
         "plot_individual": "21_measures.ipynb",
         "plot_continuous_time_domain": "21_measures.ipynb",
         "discretise_player_bets": "21_measures.ipynb",
         "plot_discrete_time_domain": "21_measures.ipynb",
         "plot_session_domain": "21_measures.ipynb",
         "plot_cost_domain": "21_measures.ipynb",
         "show_domain_differences": "21_measures.ipynb",
         "descriptive_table": "22_statistics.ipynb",
         "ks_test": "22_statistics.ipynb",
         "cohens_d": "22_statistics.ipynb",
         "spearmans_r": "22_statistics.ipynb",
         "calculate_walker_matrix": "22_statistics.ipynb",
         "label_overlap_table": "22_statistics.ipynb",
         "color_matrix": "22_statistics.ipynb",
         "top_split": "23_labelling.ipynb",
         "get_labelled_groups": "23_labelling.ipynb",
         "spending_portions": "23_labelling.ipynb",
         "k_means": "24_machine_learning.ipynb",
         "k_means_range": "24_machine_learning.ipynb",
         "k_means_ensemble": "24_machine_learning.ipynb",
         "agglomerative_cluster": "24_machine_learning.ipynb",
         "describe_clusters": "24_machine_learning.ipynb",
         "neural_network_classification": "24_machine_learning.ipynb",
         "neural_network_regression": "24_machine_learning.ipynb",
         "logistic_regression": "24_machine_learning.ipynb",
         "lasso_logistic_regression": "24_machine_learning.ipynb",
         "svm_eps_regression": "24_machine_learning.ipynb",
         "svm_c_classification": "24_machine_learning.ipynb",
         "svm_one_classification": "24_machine_learning.ipynb",
         "rf_regression": "24_machine_learning.ipynb",
         "rf_classification": "24_machine_learning.ipynb",
         "compute_performance": "24_machine_learning.ipynb",
         "plot_cluster_sizes": "24_machine_learning.ipynb",
         "plot_agglomeration_dendrogram": "24_machine_learning.ipynb"}

modules = ["data.py",
           "measures.py",
           "statistics.py",
           "labelling.py",
           "machine_learning.py",
           "__init__.py"]

doc_url = "https://gamba-dev.github.io/gamba/"

git_url = "https://github.com/gamba-dev/gamba/tree/main/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: 24_machine_learning.ipynb (unless otherwise specified).

__all__ = ['k_means', 'k_means_range', 'k_means_ensemble', 'agglomerative_cluster', 'describe_clusters',
           'neural_network_classification', 'neural_network_regression', 'logistic_regression',
           'lasso_logistic_regression', 'svm_eps_regression', 'svm_c_classification', 'svm_one_classification',
           'rf_regression', 'rf_classification', 'compute_performance', 'plot_cluster_sizes',
           'plot_agglomeration_dendrogram']

# Cell
from sklearn.cluster import KMeans
import numpy as np
#export
def k_means(measures_table, clusters=4, data_only=False, plot=False, loud=False):
    "Applies the k-means clustering algorithm to a measures table. The resulting clustering is then reported in terms of its inertia (sum of squared distances of samples to their closest cluster center) and its silhouette score (how distinct clusters are within the sample [see the skikit learn docs for details]). The measures passed as the first parameter can be returned with an added column reporting the cluster each player belongs to using the data_only parameter."

    # get variable names from the behavioural measures
    variables = list(measures_table.columns)[1:]

    Kmean = KMeans(n_clusters=clusters)

    data = np.array(measures_table[variables].values)
    Kmean.fit(data)

    silhouette = metrics.silhouette_score(data, Kmean.labels_, metric="euclidean")

    cluster_centers = Kmean.cluster_centers_

    clustered_data = measures_table.copy()
    clustered_data["cluster"] = Kmean.labels_

    if loud:
        print("variables:", variables)
        print("centers:", Kmean.cluster_centers_)
        print("inertia:", Kmean.inertia_)
        print("silhouette:", silhouette)

    if plot:
        bars = []
        heights = []
        for label in set(sorted(Kmean.labels_)):
            bars.append(label)
            heights.append(list(Kmean.labels_).count(label))

        colors = ["C0", "C1", "C2", "C3", "C4", "C5", "C6", "C7", "C8", "C9"]
        plt.bar(bars, heights, color=colors[: len(bars)])
        plt.title(
            "\nClusters: "
            + str(len(bars))
            + "\nInertia: "
            + str(round(Kmean.inertia_))
            + "\nIterations: "  # Kmean.inertia_ is the sum of squared distances of samples to their closest cluster center
            + str(Kmean.n_iter_),
            x=1.01,
            y=0.5,
            ha="left",
        )
        plt.xlabel("Cluster ID")
        plt.ylabel("Number of Members")
        plt.show()

    if data_only:
        return clustered_data

    return clustered_data, Kmean.inertia_, silhouette

# Cell
def k_means_range(measures_table, min_clusters=2, max_clusters=13):
    "Computes the k_means calculation above across a range of cluster counts, returning their goodness of fit measures (inertia and silhouette)."

    # print('calculating k means in range', min_clusters, max_clusters)

    inertias = []
    silhouettes = []

    for x in range(min_clusters, max_clusters + 1):
        k_means_result = k_means(measures_table, clusters=x)
        inertias.append(k_means_result[1])
        silhouettes.append(k_means_result[2])

    return inertias, silhouettes

# Cell
def k_means_ensemble(measures_table, ensemble_size=100, min_clusters=2, max_clusters=13):
    "Computes the k_means clustering algorithm across a range of cluster counts, a number of times. This is useful for determining clusters in a more robust way but can be slow on large data sets."

    all_inertias = []
    all_silhouettes = []

    # call the k_means_range function n times, storing scores in the above arrays
    for x in range(ensemble_size):
        k_means_range_result = k_means_range(
            measures_table, min_clusters=min_clusters, max_clusters=max_clusters
        )
        all_inertias.append(k_means_range_result[0])
        all_silhouettes.append(k_means_range_result[1])

    # now average each of the elements in the score lists
    ensemble_inertias = []
    ensemble_silhouettes = []
    for cluster_num in range(len(all_inertias[0])):
        inertia_scores = [all_inertias[x][cluster_num] for x in range(ensemble_size)]
        ensemble_inertias.append(statistics.mean(inertia_scores))

        silhouette_scores = [
            all_silhouettes[x][cluster_num] for x in range(ensemble_size)
        ]
        ensemble_silhouettes.append(statistics.mean(silhouette_scores))

    return ensemble_inertias, ensemble_silhouettes

# Cell
from sklearn.cluster import AgglomerativeClustering
#export
def agglomerative_cluster(measures_table, distance_threshold=0, n_clusters=None):
    "Performs sklearn's agglomerative clustering algorithm on a dataframe of behavioural measures. See their documentation for details. Note: Either the distance threshold or the n_cluster parameter must be None."

    variables = list(measures_table.columns)[1:]
    X = measures_table[variables].values

    model = AgglomerativeClustering(
        distance_threshold=distance_threshold, n_clusters=n_clusters
    )
    model = model.fit(X)
    return model

# Cell
import pandas as pd
def describe_clusters(clustered_measures_table, cluster_col="cluster"):
    "Describes cluster centroids (mean values of each measure) for each cluster in a clustered measures table."

    unique_clusters = list(set(clustered_measures_table["cluster"].values))

    descriptive_table = pd.DataFrame()
    descriptive_table["cluster_centroid"] = clustered_measures_table.columns[1:]

    for value in unique_clusters:
        members = clustered_measures_table[
            clustered_measures_table[cluster_col] == value
        ]
        centroid = [members[col].mean() for col in members.columns[1:]]
        descriptive_table["n=" + str(len(members))] = centroid

    descriptive_table.set_index("cluster_centroid", inplace=True)
    return descriptive_table

# Cell
from keras.layers import Dense, Activation
from keras.models import Sequential
import numpy as np
import gamba.measures as measures

def neural_network_classification(measures_table, label, train_test_split=0.7):
    "A basic example of a classification network using a single sigmoid-activated output node. The output is the probability of the player belonging to the value of the label. As with the `neural_network_regression` this network closely follows examples in [Deep Learning with Python](http://faculty.neu.edu.cn/yury/AAI/Textbook/Deep%20Learning%20with%20Python.pdf) which is a great resource for extending this example."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id',label], axis=1)
    test_labels = test_measures[label]

    model = Sequential()
    model.add(Dense(17, activation = 'relu', input_dim = 17))
    model.add(Dense(50, activation = 'relu'))
    model.add(Dense(units = 1, activation='sigmoid')) # this is the 'classification' part

    model.compile(optimizer = 'adam',
                  loss = 'mean_squared_error',
                  metrics=['accuracy'])

    history = model.fit(train_data, train_labels,
                        batch_size = 20, epochs = 100,
                        validation_data=(test_data, test_labels),
                        verbose=False)

    # now make a prediction and clip the values to 0 or 1 as in the original code
    raw_prediction = model.predict(test_data)
    predicted_labels = [value[0] for value in np.where(raw_prediction >= 0.5, 1, 0)]

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def neural_network_regression(measures_table, label, train_test_split=0.7):
    "A very simple neural network for regression using two hidden layers and one output node with a linear activation function. This method is very similar to the example in [Deep Learning with Python](http://faculty.neu.edu.cn/yury/AAI/Textbook/Deep%20Learning%20with%20Python.pdf), and can be used as a template for more complex network topologies."

    # note that the metric of accuracy doesn't apply for regression problems (mean absolute error is better)

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id',label], axis=1)
    test_labels = test_measures[label]

    model = Sequential()
    model.add(Dense(17, activation = 'relu', input_dim = 17))
    model.add(Dense(50, activation = 'relu'))
    model.add(Dense(units = 1)) # see page 86 of Deep Learning with Python by Chollet for an explanation

    model.compile(optimizer = 'adam',
                  loss = 'mean_squared_error',
                  metrics=['accuracy'])

    history = model.fit(train_data, train_labels,
                        batch_size = 20, epochs = 100,
                        validation_data=(test_data, test_labels),
                        verbose=False)

    # now make a prediction and clip the values to 0 or 1 as in the original code
    raw_prediction = model.predict(test_data)
    predicted_labels = [value[0] for value in np.where(raw_prediction >= 0.5, 1, 0)]

    return test_labels, predicted_labels

# Cell
import statsmodels.api as sm
from sklearn.linear_model import LogisticRegression
from sklearn import svm
import gamba.measures as measures
#export
def logistic_regression(measures_table, label, train_test_split=0.7):
    "Performs a logistic regression using the [statsmodels library](https://www.statsmodels.org/stable/index.html), returning the predicted labels rounded to the nearest integer. Note: this method is currently hard-coded to only function on Philander 2014's data set. Abstracted logistic regression function coming soon."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]


    # defines the R style formula to fit
    formula = str(label) + " ~ gender+age+total_wagered+num_bets+frequency+duration+bets_per_day+net_loss+intensity+variability+frequency_1m+trajectory+z_intensity+z_variability+z_frequency+z_trajectory"
    model = sm.formula.glm(formula=formula, family=sm.families.Binomial(), data=train_measures)

    # this is where the stepwise bit could happen - see original code
    fit_model = model.fit()

    raw_prediction = fit_model.predict(test_measures)
    predicted_labels = [value for value in np.where(raw_prediction >= 0.5, 1, 0)]

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def lasso_logistic_regression(measures_table, label, train_test_split=0.7):
    "Performs a 'lasso' (optimizes a least-square problem with L1 penalty) logistic regression using [sklearn's linear_model](https://scikit-learn.org/stable/modules/classes.html#module-sklearn.linear_model). This [stackoverflow post](https://stackoverflow.com/questions/41639557/how-to-perform-logistic-lasso-in-python>) contains a useful discussion on this type of function-estimation regression pair."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = LogisticRegression(penalty='l1', solver='liblinear')

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def svm_eps_regression(measures_table, label, train_test_split=0.7):
    "Creates and trains a support vector machine for epsilon-support vector regression using the sklearn library's implementation."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = svm.SVR(kernel='rbf')

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    # convert probabilities to binary labels for comparison
    regression_cutoff = 0.5
    predicted_labels = np.where(predicted_labels < regression_cutoff, 0, 1)

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def svm_c_classification(measures_table, label, train_test_split=0.7):
    "Creates and trains a support vector machine for classification using the sklearn library's implementation."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = svm.SVC(kernel='rbf')

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def svm_one_classification(measures_table, label, train_test_split=0.7):
    "Creates and trains a support vector machine for one-class classification using the sklearn library's implementation."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = svm.OneClassSVM(kernel='rbf')

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    # need to add a correction step for the labels here as OneClassSVM returns -1 for outliers and 1 for inliers
    predicted_labels = np.where(predicted_labels < 0, 1, 0)

    return test_labels, predicted_labels

# Cell
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import RandomForestClassifier
import gamba.measures as measures
#export
def rf_regression(measures_table, label, train_test_split=0.7):
    "Creates and fits a random forest regressor using [sklearn's ensemble module](https://scikit-learn.org/stable/modules/ensemble.html#forest), returning the predicted labels rounded to the nearest integer."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = RandomForestRegressor()

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    # convert probabilities to binary labels for comparison
    regression_cutoff = 0.5
    predicted_labels = np.where(predicted_labels < regression_cutoff, 0, 1)

    return test_labels, predicted_labels

# Cell
import gamba.measures as measures
def rf_classification(measures_table, label, train_test_split=0.7):
    "Creates a random forest for classification also using the sklearn library's implementation."

    train_measures, test_measures = measures.split_measures_table(measures_table, frac=train_test_split)
    test_labels = test_measures[label]

    train_data = train_measures.drop(['player_id', label], axis=1)
    train_labels = train_measures[label]
    test_data = test_measures.drop(['player_id', label], axis=1)

    model = RandomForestClassifier(n_estimators=100)

    model.fit(train_data, train_labels)

    predicted_labels = model.predict(test_data)

    return test_labels, predicted_labels

# Cell
from sklearn import metrics
#export
def compute_performance(method_name, actual, predicted):
    "Computes performance metrics including sensitivity, specificity, accuracy, confusion matrix values, odds ratio, and area under curve, for a given classification/regression using its actual and predicted values."

    # resources:
    # describes odds ratio and precision equations
    # https://cran.r-project.org/web/packages/ROCR/ROCR.pdf
    # describes sklearn's confusion matrix
    # https://scikit-learn.org/stable/modules/generated/sklearn.metrics.confusion_matrix.html#sklearn-metrics-confusion-matrix

    result = metrics.classification_report(actual, y_pred=predicted, output_dict=True)
    sensitivity = result['1']['recall']
    specificity = result['0']['recall']
    accuracy = result['accuracy']
    fpr, tpr, thresholds = metrics.roc_curve(actual, predicted)
    auc = metrics.auc(fpr, tpr)
    confusion_matrix = metrics.confusion_matrix(actual, predicted)
    tn, fp, fn, tp = confusion_matrix.ravel()

    # odds ratio is (tp x tn)/(fp x fn)
    odds_ratio = 0
    if fp != 0 and fn != 0:
        odds_ratio = (tp * tn)/(fp * fn)

    # precision is tp / (tp + fp)
    precision = tp / (tp + fp)

    metrics_df = pd.DataFrame()
    metrics_df['sensitivity'] = [round(sensitivity, 3)]
    metrics_df['specificity'] = [round(specificity, 3)]
    metrics_df['accuracy'] = [round(accuracy, 3)]
    metrics_df['precision'] = [round(precision, 3)]
    metrics_df['auc'] = [round(auc, 3)]
    metrics_df['odds_ratio'] = [round(odds_ratio, 3)]
    metrics_df.index = [method_name]
    return metrics_df

# Cell
def plot_cluster_sizes(model):
    "Create a bar chart using a previously computed clustering model. Each bar represents a single cluster, with the height of the bars representing the number of members (players) in each cluster."

    plt.figure()
    cluster_ids = list(set(list(model.labels_)))
    cluster_sizes = [list(model.labels_).count(x) for x in cluster_ids]
    plt.bar(
        cluster_ids,
        cluster_sizes,
        color=plt.rcParams["axes.prop_cycle"].by_key()["color"],
    )
    locs, labels = plt.xticks()
    plt.xticks(range(len(cluster_ids)), cluster_ids)
    plt.xlabel("Cluster ID")
    plt.ylabel("Number of Players per Cluster")
    plt.grid(axis="x")
    return plt

# Cell
def plot_agglomeration_dendrogram(model, dt_cutoff=None, **kwargs):
    "Create a dendrogram visualising a heirarchical clustering method (agglomerative clustering). A horisontal line can be added using the dt_cutoff parameter to visualise the number of clusters at a given distance threshold."

    # Create linkage matrix and then plot the sch.dendrogram
    # create the counts of samples under each node
    counts = np.zeros(model.children_.shape[0])
    n_samples = len(model.labels_)
    for i, merge in enumerate(model.children_):
        current_count = 0
        for child_idx in merge:
            if child_idx < n_samples:
                current_count += 1  # leaf node
            else:
                current_count += counts[child_idx - n_samples]
        counts[i] = current_count

    linkage_matrix = np.column_stack(
        [model.children_, model.distances_, counts]
    ).astype(float)

    # Plot the corresponding dendrogram
    plt.figure(figsize=(12, 4))
    plt.title("Hierarchical Clustering dendrogram")
    sch.dendrogram(linkage_matrix, truncate_mode="level", p=3)
    if dt_cutoff != None:
        plt.plot(list(plt.xlim()), [dt_cutoff, dt_cutoff], linestyle="--", color="grey")
    plt.xlabel("Number of points in node (or index of point if no parenthesis).")
    plt.ylabel("Distance threshold")
    plt.grid(False)
    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 20_data.ipynb (unless otherwise specified).

__all__ = ['read_csv', 'concat', 'add_tables', 'prepare_labrie_data', 'prepare_braverman_data',
           'prepare_philander_data', 'generate_bets', 'check_data', 'summarise_app', 'summarise_providers',
           'plot_player_career', 'plot_player_career_split', 'plot_provider_dates', 'plot_provider_bets']

# Cell
import pandas as pd
def read_csv(file, parse_dates=[], index_col=None, delimiter=",", dummy_data=False):
    "Read csv files into a pandas dataframe."
    df = pd.read_csv(
        file, parse_dates=parse_dates, index_col=index_col, delimiter=delimiter
    )

    return df

# Cell
def concat(dfs, axis=0):
    "This method is a wrapper of pandas' **concat** function. See [here](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.concat.html) for details."

    df = pd.concat(dfs, axis=axis)
    return df

# Cell
def add_tables(t1, t2, same_columns=False):
    "Joins two tables (the second to the right hand side of the first), adding '_2' to column names if same_columns parameter is True."

    if same_columns:
        t2.columns = [name + "_2" for name in t2.columns]
    combined = pd.concat([t1, t2.reindex(t1.index)], axis=1)
    return combined

# Cell
def prepare_labrie_data(filename, savedir="labrie_individuals/", loud=False, year=2008):
    "Splits the original labrie data into CSV files for each individual's transactions and renames the columns to be compatable with the rest of the gamba library."

    labrie_data = None

    if year == 2008:
        labrie_data = pd.read_csv(filename, delimiter="\t", parse_dates=["Date"])
    elif year == 2007:
        labrie_data = pd.read_csv(filename, parse_dates=["Date"])

    # rename columns to make them compatable with gamba.measures
    if loud:
        print("original columns:", list(labrie_data.columns))

    if year == 2008:
        labrie_data.columns = [
            "player_id",
            "bet_time",
            "bet_size",
            "payout_size",
            "bet_count",
        ]
        labrie_data.to_csv("gamba_ready_labrie_data_2008.csv", index=False)
    elif year == 2007:
        labrie_data.columns = [
            "player_id",
            "bet_time",
            "product_id",
            "bet_size",
            "payout_size",
            "bet_count",
        ]
        labrie_data.to_csv("gamba_ready_labrie_data_2007.csv", index=False)

    if loud:
        print("better columns:", list(labrie_data.columns))

    # split_individual_transactions(labrie_data, savedir)

    if loud:
        print("LaBrie data ready to use!")

    return labrie_data

# Cell
def prepare_braverman_data(filename, loud=False):
    "Splits the original Braverman and Shaffer data into CSV files for each indivdiual's transactions, and renames the columns to be compatable with the rest of the gamba library."

    braverman_data = pd.read_csv(filename, parse_dates=["TimeDATE"], delimiter="\t")

    braverman_data.columns = [
        "player_id",
        "bet_time",
        "bet_size",
        "payout_size",
        "bet_count",
    ]

    # split_individual_transactions(raw_data, 'braverman_individuals/')

    if loud:
        print("Braverman data ready to use!")

    braverman_data.to_csv("gamba_ready_braverman_data.csv", index=False)
    return braverman_data

# Cell
import numpy as np
def prepare_philander_data(filename, loud=False):
    "Loads in the analytic data set of high-risk internet gamblers and removes the UserID, Sereason, random, and clustering columns as described in Philander's 2014 study."

    analytic_data = read_csv(filename, delimiter='\t')
    philander_data = analytic_data.copy()
    philander_data['self_exclude'] = np.where(philander_data['Sereason'] == 3, 1, 0) # apply the binary self-exclude technique (middle of page 5)
    philander_data.drop(labels=['Sereason','random','p2clusteringactivity','p2clusterhalf1','p2clusterhalf2'], axis=1, inplace=True)
    philander_data.columns = ['player_id', 'country', 'gender','age','total_wagered','num_bets','frequency','duration','bets_per_day','net_loss',
                              'intensity','variability','frequency_1m','trajectory',
                              'z_intensity','z_variability','z_frequency','z_trajectory','self_exclude']
    if loud:
        print(len(philander_data), 'players loaded')

    return philander_data

# Cell
import gamba as gb, pandas as pd, random
import datetime

def generate_bets(n=200):
    df = pd.DataFrame()

    df['player_id'] = [1] * n
    df['bet_size'] = [random.randint(0,20) for _ in range(n)]
    df['bet_time'] = [datetime.datetime.now() + random.random() * datetime.timedelta(days=500) for _ in range(n)]
    df['payout_size'] = [bet * 2 if (random.random() < 0.5) else 0 for bet in df['bet_size'] ]

    df['payout_time'] = [bet_time + random.random() * datetime.timedelta(minutes=1) for bet_time in df['bet_time']]
    df['decimal_odds'] = [2.0] * n

    return df

# Cell
def check_data(dataframe):
    "Make sure that your data is in the gamba standard format (has the right column names). This method will raise an exception if the format is incorrect, and will do nothing if it is correct."

    acceptable_names = ['player_id','bet_size','bet_time','payout_size','payout_time','decimal_odds','house_edge','game_type','provider']

    for name in dataframe.columns:
        if name not in acceptable_names:
            raise Exception('invalid column name provided, expecting one of', acceptable_names)

# Cell
def summarise_app(player_bets, dapp=False):
    "Prints out some basic information about a gambling or gambling-like application given a collection of bets made through that application. Data displayed includes the number of users, the number of game types provided, the number of bets placed, the total value of the bets and the payouts, the time of the first bet, and the time of the last."

    users = set(player_bets["player_id"].values)
    games = set(player_bets["game_type"].values)
    bets = player_bets["bet_size"].sum()
    payouts = player_bets["payout_size"].sum()

    # this part is specific to gambling DApps (see Scholten 2020)
    if dapp:
        start_block = player_bets["block_number"].min()
        end_block = player_bets["block_number"].max()

    start = player_bets["bet_time"].min()
    end = player_bets["bet_time"].max()

    print("users:", len(users))
    print("games:", len(games))
    print("num bets:", len(player_bets))
    print("bet value:", bets)
    print("payout value:", payouts)

    if dapp:
        print("start:", start, int(start_block))
        print("end:", end, int(end_block))
    else:
        print("start:", start)
        print("end:", end)
    print("")

# Cell
def summarise_providers(player_bets, providers, game_types=['coinflip','onedice','twodice','roll']):
    "Create a table containing summary data for providers using a collection of player bets. Summary includes the number of unique users and games, the total value of bets and payouts, the starting and ending block numbers, and the time the starting and ending blocks occurred."

    all_summaries = pd.DataFrame()
    for provider in providers:
        provider_bets = player_bets[(player_bets['provider'] == provider) & (player_bets['game_type'].isin(game_types))]

        users = len(set(provider_bets["player_id"].values))
        games = len(set(provider_bets["game_type"].values))
        bets = provider_bets["bet_size"].sum()
        payouts = provider_bets["payout_size"].sum()
        start_block = provider_bets["block_number"].min()
        end_block = provider_bets["block_number"].max()
        start = provider_bets["bet_time"].min()
        end = provider_bets["bet_time"].max()

        summary = pd.DataFrame()
        summary['values'] = ['Unique Users', 'Games','Bet Value','Payout Value','Start Block','End Block','Start Time','End Time']
        summary[provider] = [users, games, bets, payouts, int(start_block), int(end_block), start, end]


        all_summaries['values'] = summary['values']
        all_summaries[provider] = summary[provider]

    all_summaries.set_index('values', inplace=True)
    return all_summaries

# Cell
import matplotlib.pyplot as plt
def plot_player_career(player_df, savename=None):
    "Creates a candlestick-style plot of a players betting activity over the course of their career. This works best on regularly-spaced sequential data but can also provide insight into intra-session win/loss patterns."

    plt.figure(figsize=[5, 3])
    previous_y_end = 0
    for i, bet in player_df.iterrows():
        bet_size = bet["bet_size"]
        payout_size = bet["payout_size"]
        bet_time = bet["bet_time"]
        payout_time = bet["payout_time"]

        start_y = previous_y_end
        end_y = 0

        # if bet loses
        if payout_size < bet_size:
            end_y = start_y - bet_size
            # plt.plot([2*i, 2*i + 1], [bet_size, payout_size], marker='o', color='red')
            plt.plot(
                [i, i], [start_y, end_y], marker="o", color="#d30505", markersize=12
            )
        else:
            end_y = start_y + payout_size
            # plt.plot([2*i, 2*i + 1], [bet_size, payout_size], marker='o', color='green')
            plt.plot(
                [i, i], [start_y, end_y], marker="o", color="#00B007", markersize=12
            )

        previous_y_end = end_y

    plt.xlabel(None)
    if savename != None:
        plt.savefig(savename, dpi=200, transparent=True)

    return plt

# Cell
def plot_player_career_split(player_df):
    "Plot a player's betting and payout trajectory on a single plot, with green indicating payouts (top) and red indicating bets (bottom). A cumulative value line is also plotted between the two. Note that the player_df must include both bet_size and payout_size columns."

    plt.figure()

    previous_y_end = 0
    for i, bet in player_df.iterrows():
        bet_size = bet["bet_size"]
        payout_size = bet["payout_size"]
        bet_time = bet["bet_time"]
        payout_time = bet["payout_time"]

        start_y = previous_y_end
        end_y = 0

        # if bet loses
        if payout_size < bet_size:
            end_y = start_y - bet_size
            # plt.plot([2*i, 2*i + 1], [bet_size, payout_size], marker='o', color='red')
            plt.plot([i, i], [start_y, end_y], marker="o", color="red")
        else:
            end_y = start_y + payout_size
            # plt.plot([2*i, 2*i + 1], [bet_size, payout_size], marker='o', color='green')
            plt.plot([i, i], [start_y, end_y], marker="o", color="green")

        previous_y_end = end_y

    bets = player_df["bet_size"].values
    payouts = player_df["payout_size"].values
    plt.plot(
        range(len(bets)),
        np.cumsum(-bets),
        marker="o",
        color="red",
        label="Cumulative Bets",
    )
    plt.plot(
        range(len(payouts)),
        np.cumsum(payouts),
        marker="o",
        color="green",
        label="Cumulative Payouts",
    )
    plt.legend()
    plt.xlim(0, len(bets) * 1.02)
    plt.ylim(-max([sum(bets), sum(payouts)]), max([sum(bets), sum(payouts)]))
    return plt

# Cell
def plot_provider_dates(player_bets, providers, provider_labels=None):
    """
    Visualises the start and end dates of bets from one or more providers on a stacked gantt style plot.

    """
    fig = plt.figure(figsize=[8,1.5])

    for i, provider in enumerate(providers):
        provider_bets = player_bets[(player_bets['provider'] == provider)]
        start = provider_bets["bet_time"].min()
        end = provider_bets["bet_time"].max()

        plt.plot([start, end], [i,i], label=provider)
        plt.scatter([start, end], [i,i], s=100, marker='|')

    ax = fig.axes[0]
    ax.xaxis.set_major_locator(mdates.MonthLocator([1,4,7,10])) # ticks on first day of each quarter
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%b %Y')) # format of only month (3 letter abbr)
    if provider_labels == None:
        plt.yticks([0,1,2], providers)
    else:
        plt.yticks(range(len(providers)), provider_labels)
    plt.ylim(-0.5, len(providers)-0.5)
    plt.grid(True)
    return plt

# Cell
import matplotlib.pyplot as plt, matplotlib.dates as mdates
from matplotlib.ticker import FuncFormatter
#export
def plot_provider_bets(player_bets, providers, provider_labels=None):
    "Visualises the cumulative number of bets from one or more providers."

    fig = plt.figure(figsize=[8,3])

    def millions(x, pos):
        'The two args are the value and tick position'
        return '%1.1fM' % (x*1e-6)

    formatter = FuncFormatter(millions)

    for i, provider in enumerate(providers):
        provider_bets = player_bets[(player_bets['provider'] == provider)].sort_values('bet_time')
        plt.plot(provider_bets['bet_time'], provider_bets['bet_size'].cumsum(), label=provider_labels[i])

    player_bets.sort_values('bet_time', inplace=True)
    plt.plot(player_bets['bet_time'], player_bets['bet_size'].cumsum(), label='All Applications')
    ax = fig.axes[0]
    ax.xaxis.set_major_locator(mdates.MonthLocator([1,4,7,10])) # ticks on first day of each quarter
    ax.xaxis.set_major_formatter(mdates.DateFormatter('%b %Y')) # format of only month (3 letter abbr)

    ax.yaxis.set_major_formatter(formatter)
    plt.grid(True)
    plt.ylabel('Total Value of Bets, in ETH')
    plt.legend()
    plt.savefig('provider_bets.pdf', bbox_inches='tight')
    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 22_statistics.ipynb (unless otherwise specified).

__all__ = ['descriptive_table', 'ks_test', 'cohens_d', 'spearmans_r', 'calculate_walker_matrix', 'label_overlap_table',
           'add_tables', 'color_matrix']

# Cell
import pandas as pd, numpy as np
from scipy import stats
def descriptive_table(measures_table, loud=False, extended=False):
    "Creates the first table found in LaBrie et al's 2008 paper, which presents descriptive statistics for each of the behavioural measures they calculated."

    # first pull all of the data out of the dictionary for more readable use
    # later on
    measure_names = list(measures_table.columns)[1:]

    means = []
    stds = []
    medians = []
    stats.iqrs = []
    for measure in measure_names:
        means.append(measures_table[measure].mean())
        stds.append(measures_table[measure].std())
        medians.append(measures_table[measure].median())
        stats.iqrs.append(stats.iqr(measures_table[measure].values))

    if loud:
        print("calculating descriptive statistics for LaBrie measures")

    descriptive_df = pd.DataFrame(columns=["measure", "mean", "std", "median"])

    descriptive_df["measure"] = measure_names
    descriptive_df["mean"] = means
    descriptive_df["std"] = stds
    descriptive_df["median"] = medians
    if extended:
        descriptive_df["iqr"] = stats.iqrs

    descriptive_df.set_index("measure", inplace=True)
    descriptive_df = descriptive_df.rename_axis(None)

    return descriptive_df

# Cell
def ks_test(measures_table):
    "Performs a one sample Kolmogorov-Smirnov test. This approximately indicates whether or not a collection of calculated behavioural measures are normally distributed."

    measure_names = list(measures_table.columns)[1:]

    scores = []
    pvals = []
    for measure in measure_names:
        result = stats.kstest(measures_table[measure], "norm")
        scores.append(result[0])
        pvals.append(result[1])

    ks_table = pd.DataFrame(columns=["Measure", "K-S Score", "p"])

    ks_table["Measure"] = measure_names
    ks_table["K-S Score"] = scores
    ks_table["p"] = pvals

    ks_table.set_index("Measure", inplace=True)
    ks_table.rename_axis(None, inplace=True)

    return ks_table

# Cell
import math
def cohens_d(measures_table, label):
    "Calculates Cohen's d value between the behavioural measures of two groups of players. Groups are distinguished using a label column which is either 1 (in group) or 0 (not in group). For example, the column 'in_top5' may represent whether or not a player is in the top 5 % of players by total amount wagered, and would be 1 or 0 for the top 5 and remaining 95 percent respectively."

    control_group = measures_table[measures_table[label] == 0]

    experimental_group = measures_table[measures_table[label] == 1]

    measure_names = list(measures_table.columns)[1:]

    # remove the label column (no point doing cohens d on it)
    measure_names.remove(label)

    d_results = []
    # do cohens d for each measure
    for measure in measure_names:
        control_measure = control_group[measure]
        experimental_measure = experimental_group[measure]

        control_mean = control_measure.mean()
        experimental_mean = experimental_measure.mean()

        control_sd = control_measure.std()
        experimental_sd = experimental_measure.std()

        control_n = len(control_measure)
        experimental_n = len(experimental_measure)

        top_line = ((control_n - 1) * control_sd ** 2) + (
            (experimental_n - 1) * experimental_sd ** 2
        )

        pooled_sd = math.sqrt(top_line / (control_n + experimental_n - 2))

        d = (control_mean - experimental_mean) / pooled_sd

        d_results.append(d)

    # make a nice dataframe to present the results
    d_table = pd.DataFrame(columns=["Measure", "Cohen's d"])
    d_table["Measure"] = measure_names
    d_table["Cohen's d"] = d_results

    d_table.set_index("Measure", inplace=True)
    d_table.rename_axis(None, inplace=True)

    return d_table

# Cell
def spearmans_r(measures_table, loud=False):
    "Calculates the coefficients (nonparametric Spearman's r) between a collection of behavioural measures. The upper-right diagonal of the resulting matrix is discarded (symmetric)."

    measure_names = list(measures_table.columns)[1:]

    data = []
    for column in measure_names:
        data.append(measures_table[column].values)

    labels = measure_names

    coefs = []
    p_values = []
    for toprow in data:
        for siderow in data:
            coef, p = stats.spearmanr(toprow, siderow)
            coefs.append(coef)
            p_values.append(p)

    coefs = np.array(coefs)
    # reshape as matrix
    coef_as_matrix = coefs.reshape(len(data), len(data))
    # cut off top-diagonal elements
    coef_as_matrix = np.tril(coef_as_matrix, -1)

    p_values = np.array(p_values)
    p_as_matrix = np.array(p_values).reshape(len(data), len(data))
    p_as_matrix = np.tril(p_as_matrix, -1)

    coef_df = pd.DataFrame(coef_as_matrix, columns=labels, index=labels)
    p_df = pd.DataFrame(p_as_matrix, columns=labels, index=labels)

    # now for string manipulation (get the dataframe in a more readable format)
    coef_df.replace(0, "", inplace=True)
    np.fill_diagonal(coef_df.values, "-")

    p_values = p_df.values
    results_size = len(coef_df.columns)
    clean_results = np.empty((results_size, results_size), dtype=object)
    for r, row in enumerate(coef_df.values):
        for e, element in enumerate(row):
            if element == "-":
                clean_results[r, e] = "-"
                continue
            if element == "":
                clean_results[r, e] = ""
                continue

            p = float(p_values[r, e])

            if p < 0.01:
                clean_results[r, e] = str(round(element, 2)) + "**"
            elif p < 0.05:
                clean_results[r, e] = str(round(element, 2)) + "*"
            else:
                clean_results[r, e] = round(element, 2)

    correlation_df = pd.DataFrame(
        clean_results, columns=coef_df.columns, index=coef_df.index
    )

    return correlation_df

# Cell
def calculate_walker_matrix(measures_tables, labels, measure="frequency", loud=False):
    "Performs a two sample Kolmogorov-Smirnov test between collections of measure from different games."

    data = []
    for measures_table in measures_tables:
        data.append(measures_table[measure].values)

    coefs = []
    p_values = []
    if loud:
        print("num tests:", len(data) * len(data))
    for toprow in data:
        for siderow in data:
            coef, p = stats.ks_2samp(toprow, siderow)
            coefs.append(coef)
            p_values.append(p)

    coefs = np.array(coefs)
    # reshape as matrixEOS
    coef_as_matrix = coefs.reshape(len(data), len(data))
    # cut off top-diagonal elements
    coef_as_matrix = np.tril(coef_as_matrix, -1)

    p_values = np.array(p_values)
    p_as_matrix = np.array(p_values).reshape(len(data), len(data))
    p_as_matrix = np.tril(p_as_matrix, -1)

    coef_df = pd.DataFrame(coef_as_matrix, columns=labels, index=labels)
    p_df = pd.DataFrame(p_as_matrix, columns=labels, index=labels)

    # now for string manipulation to get the dataframe in a more readable format
    coef_df.replace(0, "", inplace=True)
    np.fill_diagonal(coef_df.values, "-")

    p_values = p_df.values
    clean_results = np.empty((len(coef_df.columns), len(coef_df.columns)), dtype=object)
    for r, row in enumerate(coef_df.values):
        for e, element in enumerate(row):
            if element == "-":
                clean_results[r, e] = "-"
                continue
            if element == "":
                clean_results[r, e] = ""
                continue

            p = float(p_values[r, e])

            if p < 0.01:
                clean_results[r, e] = str(round(element, 2)) + "**"
            elif p < 0.05:
                clean_results[r, e] = str(round(element, 2)) + "*"
            else:
                clean_results[r, e] = round(element, 2)

    clean_df = pd.DataFrame(clean_results, columns=coef_df.columns, index=coef_df.index)
    return clean_df

# Cell
def label_overlap_table(measures_table, labels):
    "Calculates the number of players under a collection of labels (exclusively), and on each pair of labels (again exclusively) in the list provided. This method can be used to reproduce the final table in LaBrie et al's 2007 paper {% cite labrie2007assessing %}."

    first_diagonal_values = []
    for label in labels:
        other_labels = labels.copy()
        other_labels.remove(label)

        records_with_label = measures_table[measures_table[label] == 1]

        records_with_only_label = records_with_label.copy()
        for other_label in other_labels:
            records_with_only_label = records_with_only_label[
                records_with_only_label[other_label] == 0
            ]

        percentage = len(records_with_only_label) / len(records_with_label) * 100

        table_entry = (
            str(len(records_with_only_label)) + " (" + str(round(percentage)) + ")"
        )
        first_diagonal_values.append(table_entry)

    left_side = np.zeros((len(first_diagonal_values), len(first_diagonal_values)))
    left_side = (pd.DataFrame(left_side)).applymap(str)

    np.fill_diagonal(left_side.values, first_diagonal_values)
    left_side.index = labels
    left_side.replace("0.0", "-", inplace=True)

    only = [label + "_only" for label in labels]
    left_side.columns = only

    # end of left side (exclusive labels)

    # get pairwise combinations of labels (both indexes and label names)
    label_combinations = []
    index_combinations = []
    for index, label in enumerate(labels[:-1]):
        for inner_index, remaining_label in enumerate(labels[index + 1 :]):
            label_combinations.append([label, remaining_label])
            index_combinations.append(
                [
                    labels.index(label),
                    labels.index(label) + labels.index(remaining_label) - 1,
                ]
            )

    # get number of exclusive labels for each pairwise combination
    combination_values = []
    percentage_values = []
    for index, combination in enumerate(label_combinations):
        records_with_first = measures_table[measures_table[combination[0]] == 1]
        records_with_both = measures_table[
            (measures_table[combination[0]] == 1)
            & (measures_table[combination[1]] == 1)
        ]

        records_with_only_both = records_with_both.copy()
        other_labels = labels.copy()
        other_labels.remove(combination[0])
        other_labels.remove(combination[1])
        for other_label in other_labels:
            records_with_only_both = records_with_only_both[
                records_with_only_both[other_label] == 0
            ]

        combination_values.append(len(records_with_only_both))
        percentage_values.append(
            len(records_with_only_both) / len(records_with_first) * 100
        )

    # create and populate a matrix (to be made into a dataframe) to hold the overlap combination results
    combination_matrix = np.zeros((len(label_combinations), len(label_combinations)))
    for index, value in enumerate(combination_values):
        combination_matrix[
            index_combinations[index][0], index_combinations[index][1]
        ] = value

    # make combination matrix a dataframe and rename columns
    combination_df = pd.DataFrame(combination_matrix)
    combination_columns = []
    for label_combination in label_combinations:
        combination_columns.append(" and ".join(label_combination) + " only")

    # get the number of records which have all labels (members of all groups)
    records_meeting_all_labels = measures_table[
        measures_table[labels[0]] == 1
    ]  # get those meeting the first label
    for label in labels[1:]:
        records_meeting_all_labels = records_meeting_all_labels[
            records_meeting_all_labels[label] == 1
        ]

    combination_df = combination_df.applymap(str)
    combination_df.replace("0.0", "-", inplace=True)

    combination_df.columns = combination_columns

    # add percentage values to exclusive columns (right side)
    for index, label_combination in enumerate(index_combinations):
        value = combination_df.iloc[label_combination[0], label_combination[1]]
        try:
            combination_df.iloc[label_combination[0], label_combination[1]] = (
                str(round(float(value)))
                + " ("
                + str(round(percentage_values[index]))
                + ")"
            )
        except:
            continue

    combination_df.index = labels

    combination_df["all labels"] = (
        str(round(len(records_meeting_all_labels)))
        + " ("
        + str(
            round(
                len(records_meeting_all_labels)
                / len(measures_table[measures_table[labels[0]] == 1])
                * 100
            )
        )
        + ")"
    )

    complete_table = pd.concat([left_side, combination_df], axis=1)

    return complete_table

# Cell
def add_tables(t1, t2, same_columns=False):
    "Joins two tables (the second to the right hand side of the first), adding '_2' to column names if same_columns parameter is True."
    if same_columns:
        t2.columns = [name + "_2" for name in t2.columns]
    combined = pd.concat([t1, t2.reindex(t1.index)], axis=1)
    return combined

# Cell
import matplotlib.cm as cm
#export
def color_matrix(matrix, cmap):
    "Creates a shaded matrix based on a color map."

    results_size = len(correlations.columns)
    values = np.empty((results_size, results_size), dtype=object)
    for r, row in enumerate(correlations.values):
        for e, element in enumerate(row):
            if element == "-":
                values[r, e] = 100
                continue
            if element == "":
                values[r, e] = np.nan
                continue
            if "*" in str(element):
                value = element.replace("*", "")
                values[r, e] = float(value) * 100
            else:
                values[r, e] = element * 100

    current_cmap = cm.get_cmap(cmap)
    current_cmap.set_bad(color="white")
    plt.imshow(np.array(values).astype(np.float), cmap=current_cmap)
    plt.yticks(range(len(correlations.columns)), list(correlations.columns))
    plt.xticks(range(len(correlations.columns)), list(correlations.columns))
    plt.xticks(rotation=90)
    cbar = plt.colorbar()
    cbar.set_ticks([-100, -80, -60, -40, -20, 0, 20, 40, 60, 80, 100])
    cbar.set_ticklabels([-1, -0.8, -0.6, -0.4, -0.2, 0, 0.2, 0.4, 0.6, 0.8, 1])
    plt.ylabel("test")
    return plt


# AUTOGENERATED! DO NOT EDIT! File to edit: 21_measures.ipynb (unless otherwise specified).

__all__ = ['create_measures_table', 'duration', 'frequency', 'frequency_aggregated', 'frequency_percent',
           'number_of_bets', 'mean_bets_per_day_aggregated', 'mean_bets_per_day', 'median_bets_per_day', 'num_sessions',
           'total_play_time', 'mean_sessions_per_day', 'median_sessions_per_day', 'mean_session_duration',
           'median_session_duration', 'mean_play_time', 'median_play_time', 'mean_bets_per_day', 'median_bets_per_day',
           'num_bets_deviation_per_day', 'inactive_day_streak_variance', 'persistence', 'bet_count_trajectory',
           'mean_bets_per_session', 'median_bets_per_session', 'bets_per_session_variance', 'session_duration_variance',
           'session_duration_trajectory', 'sawtooth_occurances', 'active_day_trajectory', 'total_wagered',
           'mean_bet_size', 'median_bet_size', 'max_bet', 'min_bet', 'bet_size_range', 'bet_size_deviation',
           'trajectory', 'mean_amount_wagered_per_day', 'median_amount_wagered_per_day',
           'amount_wagered_per_day_variance', 'mean_amount_wagered_per_session', 'median_amount_wagered_per_session',
           'total_amount_wagered_across_duration', 'amount_wagered_per_session_variance', 'bet_size_trajectory',
           'net_loss', 'percent_loss', 'payout_bet_count_ratio', 'payout_bet_size_ratio', 'mean_net_loss_per_session',
           'median_net_loss_per_session', 'sum_of_payouts', 'relative_big_win', 'big_win', 'mean_payout_size',
           'median_payout_size', 'overall_binary_loser', 'clamped_net_win', 'net_loss_on_last_day', 'mean_loss_per_bet',
           'median_loss_per_bet', 'theoretical_loss', 'mean_odds_per_bet', 'median_odds_per_bet', 'num_providers_used',
           'mean_game_types_per_day', 'calculate_labrie_measures', 'calculate_braverman_measures', 'available_measures',
           'check_measure_data', 'get_sessions', 'get_daily_bets', 'standardise_measures_table', 'split_measures_table',
           'generic_trajectory', 'plot_individual', 'plot_continuous_time_domain', 'discretise_player_bets',
           'plot_discrete_time_domain', 'plot_session_domain', 'plot_cost_domain', 'show_domain_differences']

# Cell
import pandas as pd
def create_measures_table(all_player_bets, measures, daily=False):
    "Creates a measures table from a collection of bets in the gamba standard format. The second parameter 'measures' should be a list of names from the measures module. See `available_measures` or the table below for currently available options."

    # get unique players
    unique_players = list(set(all_player_bets['player_id'].values))

    # create holding structure for each of the measures
    measures_table_values = [unique_players]


    # for each of the measures
    for measure in measures:
        all_players_measures = []

        # for each of the players
        for player in unique_players:

            player_bets = all_player_bets[all_player_bets['player_id'] == player]
            # compute and add to holding structure
            this_measure = eval(measure + '(player_bets)')

            all_players_measures.append(this_measure)

        measures_table_values.append(all_players_measures)

    # turn into tidy dataframe
    measures_table = pd.DataFrame()
    measures_table['player_id'] = unique_players
    for x, measure in enumerate(measures):
        measures_table[measure] = measures_table_values[x + 1]

    return measures_table

# Cell
def duration(player_bets):
    "The number of days between the first bet and the last."
    check_measure_data(player_bets, ["bet_time"])
    player_bets.sort_values("bet_time", inplace=True)
    player_bets.reset_index(drop=True, inplace=True)

    first_day = player_bets.iloc[0]["bet_time"]
    last_day = player_bets.iloc[len(player_bets) - 1]["bet_time"]
    # add one to make it interpret as 'days where betting has occurred'
    age_in_days = (last_day.date() - first_day.date()).days + 1
    return age_in_days

# Cell
def frequency(player_bets):
    "The number of days that included at least one bet. Note that this method returns the raw number of days."
    check_measure_data(player_bets, ["bet_time"])

    player_bets.sort_values("bet_time", inplace=True)
    player_bets.reset_index(drop=True, inplace=True)

    betting_dates = player_bets["bet_time"].dt.date

    return len(betting_dates.unique())

def frequency_aggregated(player_bets):
    "Number of active betting days in first month, if data set contains daily aggregate data."

    first_day = player_bets.iloc[0]["bet_time"]
    one_month_later = first_day + datetime.timedelta(days=+30)
    first_month_bets = player_bets[player_bets["bet_time"] <= one_month_later]

    return len(first_month_bets)

# Cell
def frequency_percent(player_bets):
    "The percentage of days within the `duration` that included at least one bet. Note that this method returns the actual percentage value e.g. 75, **not** the raw number of days or values in the interval [0,1]."
    check_measure_data(player_bets, ["bet_time"])
    player_bets.sort_values("bet_time", inplace=True)
    player_bets.reset_index(drop=True, inplace=True)
    age_in_days = duration(player_bets)

    player_bets["bet_date"] = player_bets["bet_time"].dt.date
    first_day = player_bets.iloc[0]["bet_time"]

    betting_days = 0
    for day in range(age_in_days):
        current_date = (first_day + datetime.timedelta(days=day)).date()
        bets_today = player_bets.iloc[player_bets["bet_date"].values == current_date]
        if len(bets_today) != 0:
            betting_days += 1

    frequency_percentage = (betting_days / age_in_days) * 100
    return frequency_percentage

# Cell
def number_of_bets(player_bets, daily=False):
    "The total number of bets made."

    if daily:
        check_measure_data(player_bets, ["bet_count"])
        return player_bets["bet_count"].sum()

    return len(player_bets)

# Cell
def mean_bets_per_day_aggregated(player_bets):
    "Mean number of bets per active betting day in first month (intensity), if data set contains daily aggregate data."

    first_day = player_bets.iloc[0]["bet_time"]
    one_month_later = first_day + datetime.timedelta(days=+30)
    first_month_bets = player_bets[player_bets["bet_time"] <= one_month_later]
    result = first_month_bets["bet_count"].mean()
    return result

def mean_bets_per_day(player_bets):
    days = get_daily_bets(player_bets)
    daily_counts = [len(day) for day in days]
    return np.mean(daily_counts)

def median_bets_per_day(player_bets):
    days = get_daily_bets(player_bets)
    daily_counts = [len(day) for day in days]
    return np.median(daily_counts)

# Cell
def num_sessions(player_bets, session_window=20):
    "The number of gambling sessions a player played, where a session is a sequence of bets each within some time (session_window) of the previous bet."

    # no need to check_measure_data here as it's done in get_sessions
    sessions = get_sessions(player_bets, session_window=session_window)

    return len(sessions)

# Cell
import datetime
def total_play_time(player_bets, session_window=20):
    "The sum of all session lengths."
    # first used by Fiedler 2011

    # no need to check_measure_data here as it's done in get_sessions
    sessions = get_sessions(player_bets, session_window=session_window)

    play_time = datetime.timedelta(0)
    for session in sessions:
        start = session['bet_time'].min()
        end = session['bet_time'].max()
        session_length = end - start

        play_time += session_length

    return play_time

# Cell
def mean_sessions_per_day(player_bets):
    daily_bets = get_daily_bets(player_bets)

    sessions_per_day = []
    for day in daily_bets:
        sessions = get_sessions(day)
        sessions_per_day.append(len(sessions))

    return np.mean(sessions_per_day)


def median_sessions_per_day(player_bets):
    daily_bets = get_daily_bets(player_bets)

    sessions_per_day = []
    for day in daily_bets:
        sessions = get_sessions(day)
        sessions_per_day.append(len(sessions))

    return np.median(sessions_per_day)

# Cell
import numpy as np
def mean_session_duration(player_bets):
    sessions = get_sessions(player_bets)
    session_durations = [session['bet_time'].max() - session['bet_time'].min() for session in sessions]
    return np.mean(session_durations)

def median_session_duration(player_bets):
    sessions = get_sessions(player_bets)
    session_durations = [session['bet_time'].max() - session['bet_time'].min() for session in sessions]
    return np.median(session_durations)

# Cell
import numpy as np

def mean_play_time(player_bets):
    # Computes total play time per day (UTC), returning the mean of those values.
    tpt = total_play_time(player_bets)
    dur = duration(player_bets)
    return tpt/dur

def median_play_time(player_bets):
    # Computes total play time per day (UTC), returning the median of those values.
    daily_bets = get_daily_bets(player_bets)
    daily_play_times = [total_play_time(bets) for bets in daily_bets]
    return np.median(daily_play_times)

# Cell
def mean_bets_per_day(player_bets, daily=False):
    "The average (mean) number of bets made on active betting days."

    if daily:

        age_in_days = duration(player_bets)
        frequency_percentage = frequency(player_bets)

        betting_days = (frequency_percentage / 100) * age_in_days

        mean_bets_per_day = (
            number_of_bets(player_bets, daily=True) / betting_days  # taken from LaBrie 2008
        )
        return mean_bets_per_day

    # otherwise compute regular mean bets per day measure
    age_in_days = duration(player_bets)
    frequency_percentage = frequency(player_bets)

    betting_days = (frequency_percentage / 100) * age_in_days

    mean_bets_per_day = len(player_bets) / betting_days  # taken from LaBrie 2008
    return mean_bets_per_day

# Cell
import numpy as np
def median_bets_per_day(player_bets):
    "Aggregates daily bets and returns the median number of bets placed across all active betting days."

    daily_bets = aggregate_bets(player_bets, aggregation='daily')

    daily_bet_count = [len(bets) for bets in daily_bets]

    return np.median(daily_bet_count)

# Cell
import numpy as np
def num_bets_deviation_per_day(player_bets):
    "Standard deviation of the number of bets placed per active betting day."
    # used in Braverman and Shaffer 2013

    daily_bets = aggregate_bets(player_bets, aggregation='daily')

    daily_bet_count = [len(bets) for bets in daily_bets]

    return np.std(daily_bet_count)

# Cell
from datetime import timedelta
import numpy as np
from pysan.core.spells import get_spells

def inactive_day_streak_variance(player_bets):

    player_bets["bet_date"] = player_bets["bet_time"].dt.date

    active_dates = player_bets['bet_date'].unique()

    all_days = pd.date_range(player_bets['bet_date'].min(), player_bets['bet_date'].max(), freq='d')

    all_days_as_sequence_with_active_as_true = [True if (date in active_dates) else False for date in all_days]

    # the day-streak data can be framed as spells in sequence analysis for easy extraction
    spells = get_spells(all_days_as_sequence_with_active_as_true)
    inactive_spells = [spell for spell in spells if (spell[0] == False)]
    inactive_spell_lengths = [spell[1] for spell in inactive_spells]

    return np.var(inactive_spell_lengths)

# Cell
from datetime import timedelta
import numpy as np
from pysan.core.spells import get_spells

def persistence(player_bets):

    active_months = player_bets['bet_time'].dt.to_period("M").unique()

    # use daily frequency then to_period to catch full period (rather than cut off part of last month)
    all_months = pd.date_range(player_bets['bet_time'].min(), player_bets['bet_time'].max(), freq='d').to_period("M").unique()

    all_months_as_sequence_with_active_as_true = [True if (month in active_months) else False for month in all_months]

    # the day-streak data can be framed as spells in sequence analysis for easy extraction
    spells = get_spells(all_months_as_sequence_with_active_as_true)
    active_month_spells = [spell[1] for spell in spells if spell[0] == True] # true is active month

    return max(active_month_spells)

# Cell
def bet_count_trajectory(player_bets):
    days = get_daily_bets(player_bets)
    daily_counts = [len(bets) for bets in days]
    return generic_trajectory(daily_counts)

# Cell
def mean_bets_per_session(player_bets):
    sessions = get_sessions(player_bets)

    return np.mean([len(session) for session in sessions])

def median_bets_per_session(player_bets):
    sessions = get_sessions(player_bets)

    return np.median([len(session) for session in sessions])

def bets_per_session_variance(player_bets):
    sessions = get_sessions(player_bets)

    return np.var([len(session) for session in sessions])

# Cell
def session_duration_variance(player_bets):
    sessions = get_sessions(player_bets)
    session_durations = [session['bet_time'].max() - session['bet_time'].min() for session in sessions]
    return np.var(session_durations)

def session_duration_trajectory(player_bets):
    sessions = get_sessions(player_bets)
    session_durations = [session['bet_time'].max() - session['bet_time'].min() for session in sessions]
    return generic_trajectory(session_durations)

# Cell
def sawtooth_occurances(player_bets, threshold=1):

    # this algorithm is described on page 8 of Adami et al's 2013 paper
    # the threshold parameter is not well defined so defaults to 1 but is
    # changeable, and ultimately determines whether or not a change in
    # bet sizing constitutes a sawtooth occurance

    first_difference = player_bets['bet_size'].diff()

    mean = first_difference.mean()
    std = first_difference.std()

    #                                                      here is the formula from the paper
    occurances = [1 for difference in first_difference if difference > mean + threshold * std]

    return len(occurances)

# Cell
import datetime

def active_day_trajectory(player_bets, time_period_length=datetime.timedelta(days=30)):

    midpoint = player_bets['bet_time'].min() + time_period_length
    endpoint = midpoint + time_period_length

    first_period = player_bets[player_bets['bet_time'] < midpoint]
    second_period = player_bets[player_bets['bet_time'].between(midpoint, endpoint)]

    f1 = frequency(first_period) # aka number of active days
    f2 = frequency(second_period)

    return generic_trajectory([f1,f2])

# Cell
def total_wagered(player_bets):
    "The total amount wagered (sum of bet sizes)."
    check_measure_data(player_bets, ["bet_size"])
    return player_bets["bet_size"].sum()

# Cell
def mean_bet_size(player_bets, daily=False):
    "The average (mean) size of bets."
    if daily:
        check_measure_data(player_bets, ["bet_size"])
        mean_bet_size = player_bets["bet_size"].sum() / number_of_bets(player_bets, daily=True)
        return mean_bet_size

    check_measure_data(player_bets, ["bet_size"])
    return player_bets["bet_size"].mean()

# Cell
def median_bet_size(player_bets):
    "Median bet size, as used in Xuan 2009"

    check_measures_data(player_bets, ['bet_size'])

    return player_bets['bet_size'].median()

# Cell
def max_bet(player_bets):
    return player_bets['value'].max()

def min_bet(player_bets):
    return player_bets['value'].min()

def bet_size_range(player_bets):
    return max_bet(player_bets) - min_bet(player_bets)

# Cell
def bet_size_deviation(player_bets, daily=False):
    "Standard deviation of stake size in first month, if data set contains daily aggregate data."

    if daily:
        first_day = player_bets.iloc[0]["bet_time"]
        one_month_later = first_day + datetime.timedelta(days=+30)
        first_month_bets = player_bets[player_bets["bet_time"] <= one_month_later]

        return first_month_bets["bet_size"].std()

    return player_bets["bet_size"].std()

# Cell
from sklearn.linear_model import LinearRegression
import numpy as np
def trajectory(player_bets, daily=False, plot=False):
    "Gradient of a linear regression fitted to the sequence of daily aggredated bet sizes."

    if daily:
        first_day = player_bets.iloc[0]["bet_time"]
        one_month_later = first_day + datetime.timedelta(days=+30)
        first_month_bets = player_bets[player_bets["bet_time"] <= one_month_later]

        # first_month_bets['mean_bet_size'] = first_month_bets['bet_size'] / first_month_bets['bet_count']

        x = np.array(range(len(first_month_bets))).reshape((-1, 1)) + 1
        y = first_month_bets["bet_size"].values

        model = LinearRegression().fit(x, y)
        r_sq = model.score(x, y)

        model_x = x
        model_y = model.coef_ * x + model.intercept_

        if plot:
            plt.figure()
            plt.scatter(x, y)
            plt.plot(model_x, model_y, color="r")
            plt.show()

        trajectory = model.coef_[0]
        return model.coef_[0]

# Cell
def mean_amount_wagered_per_day(player_bets):
    return total_amount_wagered(player_bets) / frequency(player_bets)

def median_amount_wagered_per_day(player_bets):

    daily_bets = get_daily_bets(player_bets)

    tam_daily = [total_amount_wagered(daily) for daily in daily_bets]

    return np.median(tam_daily)

# Cell
def amount_wagered_per_day_variance(player_bets):

    daily_bets = get_daily_bets(player_bets)

    tam_daily = [total_amount_wagered(daily) for daily in daily_bets]

    return np.var(tam_daily)

# Cell
import numpy as np
def mean_amount_wagered_per_session(player_bets):

    sessions = get_sessions(player_bets)
    tam_sessions = [total_amount_wagered(session) for session in sessions]
    return np.mean(tam_sessions)

def median_amount_wagered_per_session(player_bets):

    sessions = get_sessions(player_bets)
    tam_sessions = [total_amount_wagered(session) for session in sessions]
    return np.median(tam_sessions)

# Cell
def total_amount_wagered_across_duration(player_bets):
    tam = total_amount_wagered(player_bets)
    dur = duration(player_bets)
    return tam/dur

# Cell
def amount_wagered_per_session_variance(player_bets):
    sessions = get_sessions(player_bets)
    tam_sessions = [total_amount_wagered(session) for session in sessions]
    return np.var(tam_sessions)

# Cell
def bet_size_trajectory(player_bets):
    days = get_daily_bets(player_bets)
    daily_sizes = [bets["bet_size"].sum() for bets in days]
    return generic_trajectory(daily_sizes)

# Cell
def net_loss(player_bets):
    "The net amount lost (sum of bet sizes minus sum of payout sizes). This is functionally identical to the negative of total amount won"

    check_measure_data(player_bets, ["bet_size", "payout_size"])
    net_loss_value = player_bets["bet_size"].sum() - player_bets["payout_size"].sum()
    return net_loss_value

# Cell
def percent_loss(player_bets):
    "The `net_loss` as a percentage of `total_wagered`."

    net_loss_value = net_loss(player_bets)
    total_wagered_value = total_wagered(player_bets)
    percent_loss_value = (net_loss_value / total_wagered_value) * 100
    return percent_loss_value

# Cell
def payout_bet_count_ratio(player_bets):
    bets = len(player_bets)
    payouts = len(player_bets[player_bets['payout_size'] != 0])
    return payouts / bets

def payout_bet_size_ratio(player_bets):
    bets = player_bets['bet_size'].sum()
    payouts = player_bets['payout_size'].sum()
    return payouts / bets

# Cell
def mean_net_loss_per_session(player_bets):

    sessions = get_sessions(player_bets)
    nl_sessions = [net_loss(session) for session in sessions]
    return np.mean(nl_sessions)

def median_net_loss_per_session(player_bets):

    sessions = get_sessions(player_bets)
    nl_sessions = [net_loss(session) for session in sessions]
    return np.median(nl_sessions)

# Cell
def sum_of_payouts(player_bets):
    return player_bets['payout_size'].sum()

def relative_big_win(player_bets, top):
    return 1 if max(player_bets['payout_size'] > top) else 0

def big_win(player_bets):
    # assumes payout_size is in USD already(!)
    return 1 if max(player_bets['payout_size'] > 1000) else 0

# Cell
def mean_payout_size(player_bets):
    payouts = player_bets[player_bets['payout_size'] != 0]
    return payouts['payout_size'].mean()

def median_payout_size(player_bets):
    payouts = player_bets[player_bets['payout_size'] != 0]
    return payouts['payout_size'].median()

# Cell
def overall_binary_loser(player_bets):
    return 1 if net_loss(player_bets) > 0 else 1

def clamped_net_win(player_bets):
    return max(-net_loss(player_bets), 0)

def net_loss_on_last_day(player_bets):
    days = get_daily_bets(player_bets)
    last_day = days[-1]
    return net_loss(last_day)

# Cell
def mean_loss_per_bet(player_bets):
    # additional measure not returned by literature review
    "The median amount lost per bet (bet size minus payout)."

    losses = player_bets['bet_size'] - player_bets['payout_size']

    return losses.mean()


def median_loss_per_bet(player_bets):
    # additional measure not returned by literature review
    "The median amount lost per bet (bet size minus payout)."

    losses = player_bets['bet_size'] - player_bets['payout_size']

    return losses.median()

# Cell
def theoretical_loss(player_bets):
    "The product of bet size and house advantage, also referred to as the gross gaming revenue."

    check_measure_data(player_bets, ['bet_size','house_edge'])

    unique_edges = player_bets['house_edge'].unique()

    tl_measure = 0
    for edge in unique_edges:
        bets_with_this_edge = player_bets[player_bets['house_edge'] == edge]

        tl_measure += bets_with_this_edge['bet_size'].sum() * (edge / 100)

    return tl_measure

# Cell
def mean_odds_per_bet(player_bets):
    return player_bets['decimal_odds'].mean()

def median_odds_per_bet(player_bets):
    return player_bets['decimal_odds'].median()

# Cell
def num_providers_used(player_bets): # regarding operators, not game types
    # this measure is technically at the account-leve/
    "The number of different gambling game providers used by the player."

    check_measures_data(player_bets, ['provider'])

    return len(player_bets['provider'].unique())

# Cell
import numpy as np
def mean_game_types_per_day(player_bets):
    "The mean number of unique game types played per active betting day. Note this does not account for the number of each game type played, only that they have been played at least once."
    # this measure is also technically at the account level

    daily_bets = aggregate_bets(player_bets, aggregation='daily')

    daily_game_types = [len(bets['game_type'].unique()) for bets in daily_bets]

    return np.mean(daily_game_types)

# Cell
from tqdm import tqdm
import datetime, pandas as pd
def calculate_labrie_measures(all_player_bets, savedir="", filename="gamba_labrie_measures.csv", loud=False, daily=True,):
    "Calculates the set of measures described in LaBrie et al's work in 2008 on casino gamblers. These measures include the durations, frequencies, number of bets, bets per day, value per bet (eth), total amount wagered, net loss, and percent loss for each player. As this method sits in the studies module, it accepts a list of dataframes representing each player's bets as input. By default, this method saves the resulting dataframe of each player's measures to 'gamba_labrie_measures.csv'. Be advised: this method can take some time for large numbers of players, the 'loud' parameter can be set to True to print out updates every 200 players."

    pd.options.mode.chained_assignment = None # supress setwithcopy warnings

    player_id = []
    all_duration = []
    all_frequency = []
    all_number_of_bets = []
    all_mean_bets_per_day = []
    all_mean_bet_size = []
    all_total_wagered = []
    all_net_loss = []
    all_percent_loss = []

    unique_players = list(set(all_player_bets["player_id"]))

    for i in tqdm(range(len(unique_players))):
        player_bets = all_player_bets[all_player_bets["player_id"] == unique_players[i]]
        player_id.append(player_bets.iloc[0]["player_id"])
        all_duration.append(duration(player_bets))
        all_frequency.append(frequency(player_bets))
        if daily:
            all_number_of_bets.append(number_of_bets(player_bets, daily=True))
            all_mean_bets_per_day.append(mean_bets_per_day(player_bets, daily=True))
            all_mean_bet_size.append(mean_bet_size(player_bets, daily=True))
        else:
            all_number_of_bets.append(number_of_bets(player_bets))
            all_mean_bets_per_day.append(mean_bets_per_day(player_bets))
            all_mean_bet_size.append(mean_bet_size(player_bets))
        all_total_wagered.append(total_wagered(player_bets))
        all_net_loss.append(net_loss(player_bets))
        all_percent_loss.append(percent_loss(player_bets))

    labrie_dict = {
        "player_id": player_id,
        "duration": all_duration,
        "frequency": all_frequency,
        "num_bets": all_number_of_bets,
        "mean_bets_per_day": all_mean_bets_per_day,
        "mean_bet_size": all_mean_bet_size,
        "total_wagered": all_total_wagered,
        "net_loss": all_net_loss,
        "percent_loss": all_percent_loss,
    }

    labrie_measures = pd.DataFrame.from_dict(labrie_dict)
    labrie_measures.to_csv(savedir + filename, index=False)

    if loud:
        print("LaBrie measures saved")

    return labrie_measures

# Cell
def calculate_braverman_measures(all_player_bets, savedir="", loud=False):
    "Calculates the set of measures described in Braverman and Shaffer's work in 2010 on high risk internet gamblers. These measures include the frequency, intensity, variability, and trajectories of each player. As this method sits in the studies module, it accepts a list of dataframes representing each player's bets as input. By default, this method saves the resulting dataframe of each player's measures to 'gamba_braverman_measures.csv'."

    pd.options.mode.chained_assignment = None # supress setwithcopy warnings

    player_id = []

    intensities = []
    variabilities = []
    frequencies = []
    trajectories = []

    sum_of_stakes = []
    total_num_bets = []
    mean_bet_sizes = []
    durations = []
    net_losses = []

    unique_players = list(set(all_player_bets["player_id"]))

    for i in tqdm(range(len(unique_players))):
        player_bets = all_player_bets[all_player_bets["player_id"] == unique_players[i]]

        player_id.append(player_bets.iloc[0]["player_id"])
        intensities.append(intensity(player_bets))
        frequencies.append(frequency(player_bets))
        variabilities.append(variability(player_bets, daily=True))
        trajectories.append(trajectory(player_bets, daily=True))

        sum_of_stakes.append(player_bets["bet_size"].sum())
        total_num_bets.append(player_bets["bet_count"].sum())
        mean_bet_sizes.append(
            player_bets["bet_size"].sum() / player_bets["bet_count"].sum()
        )
        durations.append(duration(player_bets))
        net_losses.append(net_loss(player_bets))

    braverman_dict = {
        "player_id": player_id,
        "intensity": intensities,
        "frequency": frequencies,
        "variability": variabilities,
        "trajectory": trajectories,
        "sum_of_stakes": sum_of_stakes,
        "total_num_bets": total_num_bets,
        "mean_bet_size": mean_bet_sizes,
        "duration": durations,
        "net_loss": net_losses,
    }

    braverman_measures = pd.DataFrame.from_dict(braverman_dict)
    braverman_measures.to_csv(savedir + "gamba_braverman_measures.csv", index=False)

    return braverman_measures

# Cell
def available_measures():
    "Returns a list of valid measures that the gamba library can compute. This list can be used in whole or in part by the `create_measures_table` method to compute these measures across one or more players bets."
    return ['duration','total_wagered','net_loss','theoretical_loss','frequency','mean_bet_size','percent_loss','mean_odds_per_bet','frequency_percent','median_bet_size','payout_bet_count_ratio','median_odds_per_bet','number_of_bets','max_bet','payout_bet_size_ratio','num_providers_used','mean_bets_per_day','min_bet','mean_net_loss_per_session','mean_game_types_per_day','median_bets_per_day','bet_size_range','median_net_loss_per_session','num_sessions','bet_size_deviation','sum_of_payouts','total_play_time','trajectory','relative_big_win','mean_sessions_per_day','mean_amount_wagered_per_day','big_win','median_sessions_per_day','median_amount_wagered_per_day','mean_payout_size','mean_play_time','amount_wagered_per_day_variance','median_payout_size','median_play_time','mean_amount_wagered_per_session','overall_binary_loser','mean_bets_per_day','median_amount_wagered_per_session','clamped_net_win','median_bets_per_day','total_amount_wagered_across_duration','net_loss_on_last_day','num_bets_deviation_per_day','amount_wagered_per_session_variance','mean_loss_per_bet','inactive_day_streak_variance','bet_size_trajectory','median_loss_per_bet','persistence','bet_count_trajectory','mean_bets_per_session','median_bets_per_session','bets_per_session_variance','session_duration_variance','session_duration_trajectory','sawtooth_occurances']

# Cell
def check_measure_data(player_bets, required_columns):
    "Compares the columns found in a dataframe of player bets to a supplied list of column names. If any of the required_column names are not found, an exception is raised reporting the error."

    for column in required_columns:
        if column not in player_bets.columns:
            exception_string = (
                "Column '" + column + "' missing from provided dataframe."
            )
            raise Exception(exception_string)

# Cell
import datetime
def get_sessions(player_bets, session_window=30):
    "Splits a single players bet dataframe into a collection of dataframes, each representing a single session. The session_window parameter determines the maximum number of minutes between bets for them to be considered part of the same session."
    check_measure_data(player_bets, ['bet_time'])

    elapsed_times = player_bets['bet_time'].diff()

    session_starts = [True if x> datetime.timedelta(minutes=session_window) else False for x in elapsed_times]

    session_ids = []
    current_session = 1
    for x in session_starts:
        if x:
            current_session += 1
        session_ids.append(current_session)

    player_bets['session_id'] = session_ids

    sessions = [x for _, x in player_bets.groupby(player_bets['session_id'])]

    return sessions

# Cell
def get_daily_bets(player_bets):
    # split player bets into multiple dataframes, each corresponding to bets placed on a single date (UTC)

    check_measure_data(player_bets, ["bet_time"])
    player_bets.sort_values("bet_time", inplace=True)
    player_bets.reset_index(drop=True, inplace=True)
    age_in_days = duration(player_bets)

    player_bets["bet_date"] = player_bets["bet_time"].dt.date
    first_day = player_bets.iloc[0]["bet_time"]

    daily_bets = []
    for day in range(age_in_days):
        current_date = (first_day + datetime.timedelta(days=day)).date()
        bets_today = player_bets[player_bets["bet_date"] == current_date]
        if len(bets_today) != 0:
            daily_bets.append(bets_today)

    return daily_bets

# Cell
import scipy.stats
def standardise_measures_table(measures_table):
    "Standardises all measures columns in a measures table by applying the scipy.stats.zscore function to each column. This is useful for column-wise comparisons and some clustering methods, but use with caution!"

    colnames = list(measures_table.columns)[1:]

    standardised_table = pd.DataFrame()
    standardised_table["player_id"] = measures_table["player_id"].values
    for col in colnames:
        standardised_table[col] = scipy.stats.zscore(measures_table[col].values)

    return standardised_table

# Cell
def split_measures_table(measures_table, frac=0.7, loud=False):
    "Splits a measures table into two randomly selected groups. This is useful for machine learning methods where a train-test split is needed, and uses the Pandas library's sample method."

    measures_table.drop(['player_id'], axis=1)
    train_table = measures_table.sample(frac=frac)
    test_table = measures_table.drop(train_table.index)

    if loud:
        print('train:test\n', len(train_table),':',len(test_table), 'ready')

    return train_table, test_table

# Cell
from sklearn.linear_model import LinearRegression
import matplotlib.pyplot as plt, numpy as np
def generic_trajectory(sequence, plot=False):

    "Gradient of a linear regression fitted to a one dimensional sequence. Generic utility method to be used in more specific cases."

    x = np.array(range(len(sequence))).reshape((-1, 1)) + 1
    y = sequence

    model = LinearRegression().fit(x, y)

    model_x = x
    model_y = model.coef_ * x + model.intercept_

    if plot:
        plt.figure()
        plt.scatter(x, y)
        plt.plot(model_x, model_y, color="r")
        plt.show()

    return model.coef_[0]

# Cell
import matplotlib.pyplot as plt
#export
def plot_individual(measures_table, player_id):
    "Plots an individual's position in the population across each of the behavioural measures in the measures table. Note that this will output one bar chart for each behavioural measure."

    measure_cols = measures_table.columns[1:]
    player = measures_table[measures_table['player_id'] == player_id]

    for measure in measure_cols:
        plt.figure(figsize=[4,2])
        hist = plt.hist(measures_table[measure].values, alpha=0.5)
        plt.plot([player[measure].values[0], player[measure].values[0]], [0, hist[0].max()], label=measure, color='black')
        plt.legend()
        plt.show()

# Cell
import matplotlib.pyplot as plt
import math
import pandas as pd
import gamba as gb
import datetime
from datetime import timedelta

def plot_continuous_time_domain(player_bets):

    plt.figure(figsize=[8,1])
    plt.scatter(player_bets['bet_time'].values, [1 for x in player_bets['bet_time']], s=100)

    return plt

def discretise_player_bets(player_bets, days=7):

    days_betting = player_bets['bet_time'].max() - player_bets['bet_time'].min()

    window_starts = [player_bets['bet_time'].min() + timedelta(days=days * x) for x in range(math.ceil(days_betting.days / days + 1))]

    discrete_blocks = [player_bets[player_bets['bet_time'].between(start, start + timedelta(days=days))] for start in window_starts]

    return discrete_blocks

def plot_discrete_time_domain(player_bets, days=7):
    days_betting = player_bets['bet_time'].max() - player_bets['bet_time'].min()

    window_starts = [player_bets['bet_time'].min() + timedelta(days=days * x) for x in range(math.ceil(days_betting.days / days + 1))]

    discrete_blocks = discretise_player_bets(player_bets)

    plt.figure(figsize=[8,1])
    for block in discrete_blocks:
        plt.scatter(block['bet_time'].values, [1 for x in block['bet_time']], s=100)

    return plt

def plot_session_domain(player_bets, session_window=30):

    sessions = get_sessions(player_bets, session_window=session_window)

    plt.figure(figsize=[8,1])
    for session in sessions:
        plt.scatter(session['bet_time'].values, [1 for x in session['bet_time']], s=100)

    return plt

def plot_cost_domain(player_bets):

    bar_y = sorted(player_bets['bet_size'].unique())
    bar_x = [len(player_bets[player_bets['bet_size'] == size]) for size in bar_y]

    plt.figure()
    plt.barh(bar_y, bar_x)
    plt.ylim(min(bar_y) - 1, max(bar_y) + 1)

    plt.show()

    plot = player_bets['bet_size'].plot.kde()
    print(type(plot))

def show_domain_differences():
    start = datetime.datetime.strptime('10/10/11', "%m/%d/%y")
    player_bets = pd.DataFrame()
    bet_times = [2,3,5,6,7, 13,14,15,16,17,18,20, 26,27,28,29,30,31,32, 39,40,41,43]
    bet_days = [start + timedelta(days=x) for x in bet_times]
    bet_sizes = [2,2,3,3,1,5,4,5,3,3,4,5,7,6,8,7,7,6,7,4,4,3,4]

    player_bets['bet_time'] = bet_days
    player_bets['bet_size'] = bet_sizes

    plot_cost_domain(player_bets)

    plot_continuous_time_domain(player_bets)

    plot_discrete_time_domain(player_bets)

    plot_session_domain(player_bets, session_window=60*48)

#show_domain_differences()


# AUTOGENERATED! DO NOT EDIT! File to edit: 23_labelling.ipynb (unless otherwise specified).

__all__ = ['top_split', 'get_labelled_groups', 'spending_portions']

# Cell
import numpy as np
def top_split(measures_table, split_by, percentile=95, loud=False):
    "Labels player's behavioural measures according to their presence in the top percentile of a given measure (split_by). Label column will be 1 if in the top percentile, 0 if not. E.g. if split_by is 'total_wagered' and percentile is 95, players in the top 5% by total wagered will be labelled 1, and all others 0."

    cutoff = np.percentile(measures_table[split_by].values, percentile)

    column_name = "top_" + split_by

    labelled_measures_table = measures_table.copy()

    labelled_measures_table[column_name] = 0
    labelled_measures_table.loc[
        labelled_measures_table[split_by] > cutoff, column_name
    ] = 1

    if loud:
        print(
            "top count:",
            len(
                labelled_measures_table[labelled_measures_table["top_" + split_by] == 1]
            ),
        )
        print(
            "other count:",
            len(
                labelled_measures_table[labelled_measures_table["top_" + split_by] == 0]
            ),
        )

    return labelled_measures_table

# Cell
def get_labelled_groups(labelled_measures_table, labelname):
    "Provides a simple way of splitting a labelled measures table into multiple tables each corresponding to a given label."

    # get the labels IN ASCENDING ORDER (important)
    labels = sorted(set(list(labelled_measures_table[labelname].values)))
    player_groups = []
    for label in labels:
        this_group = labelled_measures_table[
            labelled_measures_table[labelname] == label
        ]
        player_groups.append(this_group)

    return player_groups

# Cell
def spending_portions(measures_table):
    "Computes the percentages of total amount wagered of groups of players by presence in the top percentages of the total amount wagered. E.g. how much does the top 5% of players by total amount wagered account for out of the total amount wagered by everyone. It currently computes this statement for the percentages {1,5,10,25,50}, and prints the statements to the console."

    percentages = [1, 5, 10, 25, 50]

    for percentage in percentages:

        labelled_measures_table = gb.top_split(
            mt, "total_wagered", percentile=100 - percentage
        )
        groups = gb.get_labelled_groups(labelled_measures_table, "top_total_wagered")
        label_members = groups[1]
        non_members = groups[0]

        total_amount = labelled_measures_table["total_wagered"].sum()

        top_one_amount = label_members["total_wagered"].sum()

        print(
            "top",
            percentage,
            "% of players account for",
            str(round(top_one_amount, 0)),
            "/",
            str(round(total_amount, 0)),
            "or",
            str(round(top_one_amount / total_amount * 100, 2)),
            "of the amount wagered.",
        )